/**
 * @file
 *
 * @brief This is my submission for Coursework Challenge 1 of UWE's Internet of
 * Things module.
 *
 * @details This a game called "Blocks", a clone of a popular and well-known
 * block-stacking game which shall not be named.
 * Blocks fall from the top of the screen and the player has to try and arrange
 * them in the most space-efficient way possible without letting the blocks
 * stack up all the way to the top of the screen.
 * @details In this version, A is used to shift the current block left and B is
 * used to shift it right.
 *
 * @author Joshua Saxby <Joshua2.Saxby@live.uwe.ac.uk>
 * @date 2020
 *
 * Student Number 18018052
 *
 * @copyright Copyright (C) Joshua Saxby 2020
 *
 * @copyright
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */

#include "MicroBit.h"


/*
 * NOTE: this global variable is our MicroBit instance.
 *
 * I'd much rather declare this in main() only (I don't like globals) and
 * interact with it from my classes via a reference or a pointer to it, however
 * this does not appear to work, for some unknown reason...
 */
MicroBit ubit;


/**
 * @brief An (X,Y) tuple type for storing 2D Cartesian quantities
 */
struct Tuple {
    int x;
    int y;

    /**
     * @brief Default constructor, zero-initialises members
     */
    Tuple()
        : Tuple(0, 0)
    {}

    /**
     * @brief Simple (X, Y) Constructor
     */
    Tuple(int x, int y)
        : x(x)
        , y(y)
    {}

    /**
     * @brief Overloaded addition operator to allow adding tuples element-wise
     */
    Tuple operator+(Tuple const &other) {
        Tuple result;
        result.x = this->x + other.x;
        result.y = this->y + other.y;
        return result;
    }

    /**
     * @brief Overloaded subtraction operator to allow subtracting tuples
     * element-wise
     */
    Tuple operator-(Tuple const &other) {
        Tuple result;
        result.x = this->x - other.x;
        result.y = this->y - other.y;
        return result;
    }
};


/**
 * @brief Points are just Tuples
 */
typedef Tuple Point;


/**
 * @brief Vectors are just Tuples
 */
typedef Tuple Vector;


/**
 * @brief The different kinds of shapes our Blocks can take
 */
enum class Shape : uint8_t {
    FLAT_L,        // looks like ¬ when facing SOUTH
    TALL_L,        // looks like L when facing NORTH
    UPSIDE_DOWN_T, // looks like T when facing SOUTH
    PIPE,          // looks like | when facing EAST or WEST
    SQUARE,        // a 2x2 square
};

const size_t SHAPE_TYPES_COUNT = 5u;

/**
 * @brief Describes rotation of the block based on compass direction
 * @details The top of the shape is said to be pointing in the given direction.
 * @note Special magic values are chosen to allow modulo-arithmetic to produce
 * easy rotation calculations
 */
enum class RotationDirection : uint8_t {
    NORTH = 0u,
    EAST = 1u,
    SOUTH = 2u,
    WEST = 3u,
};

/*
 * we store the sprites for our different block shapes in FLASH, i.e. read-only
 * storage. This requires manual data-packing as follows.
 * For most of them, we store four different orientations.
 */
const uint8_t FLAT_L_NORTH_SPRITE[] __attribute__ ((aligned (4))) = {
    0xff, 0xff, // magic number requesting data to be stored in FLASH
    3, 0, // image width followed by zero
    3, 0, // image height followed by zero
    // image data follows:
    1, 0, 0,
    1, 1, 1,
    0, 0, 0,
};

const uint8_t FLAT_L_EAST_SPRITE[] __attribute__ ((aligned (4))) = {
    0xff, 0xff, // magic number requesting data to be stored in FLASH
    3, 0, // image width followed by zero
    3, 0, // image height followed by zero
    // image data follows:
    0, 1, 1,
    0, 1, 0,
    0, 1, 0,
};

const uint8_t FLAT_L_SOUTH_SPRITE[] __attribute__ ((aligned (4))) = {
    0xff, 0xff, // magic number requesting data to be stored in FLASH
    3, 0, // image width followed by zero
    3, 0, // image height followed by zero
    // image data follows:
    0, 0, 0,
    1, 1, 1,
    0, 0, 1,
};

const uint8_t FLAT_L_WEST_SPRITE[] __attribute__ ((aligned (4))) = {
    0xff, 0xff, // magic number requesting data to be stored in FLASH
    3, 0, // image width followed by zero
    3, 0, // image height followed by zero
    // image data follows:
    0, 1, 0,
    0, 1, 0,
    1, 1, 0,
};

const uint8_t TALL_L_NORTH_SPRITE[] __attribute__ ((aligned (4))) = {
    0xff, 0xff, // magic number requesting data to be stored in FLASH
    3, 0, // image width followed by zero
    3, 0, // image height followed by zero
    // image data follows:
    0, 1, 0,
    0, 1, 0,
    0, 1, 1,
};

const uint8_t TALL_L_EAST_SPRITE[] __attribute__ ((aligned (4))) = {
    0xff, 0xff, // magic number requesting data to be stored in FLASH
    3, 0, // image width followed by zero
    3, 0, // image height followed by zero
    // image data follows:
    0, 0, 0,
    1, 1, 1,
    1, 0, 0,
};

const uint8_t TALL_L_SOUTH_SPRITE[] __attribute__ ((aligned (4))) = {
    0xff, 0xff, // magic number requesting data to be stored in FLASH
    3, 0, // image width followed by zero
    3, 0, // image height followed by zero
    // image data follows:
    1, 1, 0,
    0, 1, 0,
    0, 1, 0,
};

const uint8_t TALL_L_WEST_SPRITE[] __attribute__ ((aligned (4))) = {
    0xff, 0xff, // magic number requesting data to be stored in FLASH
    3, 0, // image width followed by zero
    3, 0, // image height followed by zero
    // image data follows:
    0, 0, 1,
    1, 1, 1,
    0, 0, 0,
};

const uint8_t UPSIDE_DOWN_T_NORTH_SPRITE[] __attribute__ ((aligned (4))) = {
    0xff, 0xff, // magic number requesting data to be stored in FLASH
    3, 0, // image width followed by zero
    3, 0, // image height followed by zero
    // image data follows:
    0, 1, 0,
    1, 1, 1,
    0, 0, 0,
};

const uint8_t UPSIDE_DOWN_T_EAST_SPRITE[] __attribute__ ((aligned (4))) = {
    0xff, 0xff, // magic number requesting data to be stored in FLASH
    3, 0, // image width followed by zero
    3, 0, // image height followed by zero
    // image data follows:
    0, 1, 0,
    0, 1, 1,
    0, 1, 0,
};

const uint8_t UPSIDE_DOWN_T_SOUTH_SPRITE[] __attribute__ ((aligned (4))) = {
    0xff, 0xff, // magic number requesting data to be stored in FLASH
    3, 0, // image width followed by zero
    3, 0, // image height followed by zero
    // image data follows:
    0, 0, 0,
    1, 1, 1,
    0, 1, 0,
};

const uint8_t UPSIDE_DOWN_T_WEST_SPRITE[] __attribute__ ((aligned (4))) = {
    0xff, 0xff, // magic number requesting data to be stored in FLASH
    3, 0, // image width followed by zero
    3, 0, // image height followed by zero
    // image data follows:
    0, 1, 0,
    1, 1, 0,
    0, 1, 0,
};

const uint8_t SQUARE_SPRITE[] __attribute__ ((aligned (4))) = {
    0xff, 0xff, // magic number requesting data to be stored in FLASH
    2, 0, // image width followed by zero
    2, 0, // image height followed by zero
    // image data follows:
    1, 1,
    1, 1,
};

const uint8_t PIPE_HORIZONTAL_SPRITE[] __attribute__ ((aligned (4))) = {
    0xff, 0xff, // magic number requesting data to be stored in FLASH
    3, 0, // image width followed by zero
    3, 0, // image height followed by zero
    // image data follows:
    0, 0, 0,
    1, 1, 1,
    0, 0, 0,
};

const uint8_t PIPE_VERTICAL_SPRITE[] __attribute__ ((aligned (4))) = {
    0xff, 0xff, // magic number requesting data to be stored in FLASH
    3, 0, // image width followed by zero
    3, 0, // image height followed by zero
    // image data follows:
    0, 1, 0,
    0, 1, 0,
    0, 1, 0,
};

/*
 * this lookup table maps shapes and directions to specific sprite images
 * NOTE: this table only handles shapes which have different sprites for every
 * direction, which is almost all of them, but does not include Square or Pipe.
 */
MicroBitImage SPRITE_SHEET[SHAPE_TYPES_COUNT][4] = {
    // FLAT_L: N, E, S, W
    {
        MicroBitImage((ImageData*)FLAT_L_NORTH_SPRITE),
        MicroBitImage((ImageData*)FLAT_L_EAST_SPRITE),
        MicroBitImage((ImageData*)FLAT_L_SOUTH_SPRITE),
        MicroBitImage((ImageData*)FLAT_L_WEST_SPRITE),
    },
    // TALL_L: N, E, S, W
    {
        MicroBitImage((ImageData*)TALL_L_NORTH_SPRITE),
        MicroBitImage((ImageData*)TALL_L_EAST_SPRITE),
        MicroBitImage((ImageData*)TALL_L_SOUTH_SPRITE),
        MicroBitImage((ImageData*)TALL_L_WEST_SPRITE),
    },
    // UPSIDE_DOWN_T: N, E, S, W
    {
        MicroBitImage((ImageData*)UPSIDE_DOWN_T_NORTH_SPRITE),
        MicroBitImage((ImageData*)UPSIDE_DOWN_T_EAST_SPRITE),
        MicroBitImage((ImageData*)UPSIDE_DOWN_T_SOUTH_SPRITE),
        MicroBitImage((ImageData*)UPSIDE_DOWN_T_WEST_SPRITE),
    },
};


/**
 * @brief Stores the shape and colour information for a Block in the game.
 */
class Block {
public:
    /**
     * @brief Default constructor, randomly selects attributes for this Block.
     */
    Block();

    /**
     * @brief Returns a read-only image of what this Block looks like
     */
    MicroBitImage image() const;

    /**
     * @brief Changes this Block to face in the given direction
     * @note No checks are made to see if this causes a collision or not
     */
    void face(RotationDirection direction);
    /**
     * @brief The direction the top of this Block is facing
     */
    RotationDirection facing;
private:
    /**
     * @brief The Shape that this Block is
     */
    Shape shape;
};

Block::Block()
    : facing((RotationDirection)ubit.random(4)) // choose a random direction
    , shape((Shape)ubit.random(SHAPE_TYPES_COUNT)) // choose a random shape
{}

MicroBitImage Block::image() const {
    // retrieve the correct image sprite
    switch (this->shape) {
    case Shape::SQUARE:
        // Squares always look the same regardless of which direction they face
        return MicroBitImage((ImageData*)SQUARE_SPRITE);
    case Shape::PIPE:
        // we only need to check if it's horizontal or vertical if it's Pipe
        return ((unsigned int)this->facing % 2u) == 0u ?
            MicroBitImage((ImageData*)PIPE_HORIZONTAL_SPRITE) : // North/South
            MicroBitImage((ImageData*)PIPE_VERTICAL_SPRITE); // East/West
    default:
        // all other types (Flat L, Tall L, Upside Down T) use the lookup table
        return SPRITE_SHEET[(uint8_t)this->shape][(uint8_t)this->facing];
    }
}


/**
 * @brief Uses collision-detection to decide whether the given Block can move
 * along the provided translation Vector.
 * @param shape The Block which is to be moved
 * @param origin The position of this Block (X,Y of top left corner)
 * @param stacked Image containing already-stacked Block pixels
 * @param move The translation Vector that is being attempted to move the Block
 */
bool can_shape_move(
    Block& shape, Point origin, MicroBitImage& stacked, Vector move
) {
    MicroBitImage shape_image = shape.image();
    // iterate over all the pixels in the image
    int width = shape_image.getWidth();
    int height = shape_image.getHeight();
    // optimisation: all our translated destinations include origin + move
    origin = origin + move;
    // optimisation: collisions more likely from bottom, search there first
    for (int y = height - 1; y >= 0; y--) {
        for (int x = 0; x < width; x++) {
            // if this pixel is not clear
            if (shape_image.getPixelValue(x, y) != 0) {
                // calculate the translated position, including origin
                Point destination = origin + Point(x, y);
                // if out-of-bounds, of course it can't move
                if (
                    destination.x < 0 or
                    destination.x > 4 or
                    destination.y > 4
                    /*
                     * NOTE: there is no check for negative Y since we allow
                     * shapes to poke out of the top of the screen as they
                     * descend into play.
                     */
                ) {
                    return false;
                }
                // if Y is negative, we don't need to check this pixel
                if (destination.y < 0) {
                    continue; // no point checking this pixel, out of screen
                }
                // check if the translated position of this pixel is not free
                if (stacked.getPixelValue(destination.x, destination.y) != 0) {
                    return false; // can't move because space is obstructed
                }
            }
        }
    }
    // otherwise, if we found no problems then it's ok
    return true;
}


/**
 * @brief Determines whether a Block can be rotated to the given position
 * without colliding with the stacked blocks.
 * @param shape The Block which is to be moved
 * @param origin The position of this Block (X,Y of top left corner)
 * @param stacked Image containing already-stacked Block pixels
 * @param direction The new direction that the Block is desired to face
 */
bool can_shape_rotate(
    Block shape,
    Point origin,
    MicroBitImage& stacked,
    RotationDirection direction
) {
    // because shape is a copy, we can just rotate that and check its image
    shape.facing = direction;
    MicroBitImage shape_image = shape.image();
    // iterate over all the pixels in the image
    int width = shape_image.getWidth();
    int height = shape_image.getHeight();
    for (int y = 0; y < height; y++) {
        for (int x = 0; x < width; x++) {
            // if this pixel is not clear
            if (shape_image.getPixelValue(x, y) != 0) {
                // calculate where this pixel is once origin translation applied
                Point destination = origin + Point(x, y);
                // if Y is negative, we don't need to check this row
                if (destination.y < 0) {
                    break; // no point checking this whole row, out of screen
                }
                // check if this pixel position is free in the stacked image
                if (stacked.getPixelValue(destination.x, destination.y) != 0) {
                    return false; // space is obstructed so can't rotate
                }
            }
        }
    }
    // otherwise, if we found no problems then it's ok
    return true;
}


/* begin game state variables */

// this image contains all stacked shapes --it covers the entire screen
MicroBitImage stacked_shapes(5, 5);
bool game_over;
// player left/right shift request tracker
int move_command = 0; // 0 = do nothing, -1 = shift left, +1 = shift right
// player clockwise/anticlockwise rotate request tracker
int rotate_command = 0; // 0 = do nothing, -1 = rotate acw, +1 = rotate cw
// player's score
int score = 0;

/* end game state variables */


/*
 * our event handler functions take the name of the in-game command they
 * represent, the binding from sensors to these is handled when they are bound
 * to particular events in start_new_game().
 */
void move_left(MicroBitEvent) {
    move_command = -1;
}


void move_right(MicroBitEvent) {
    move_command = +1;
}


void rotate_acw(MicroBitEvent) {
    rotate_command = -1;
}


void rotate_cw(MicroBitEvent) {
    rotate_command = +1;
}


/**
 * @brief Returns true/false for whether the given row of the given image is
 * contiguous or not, that is, if it is full of non-clear pixels.
 */
bool is_row_contiquous(MicroBitImage stacked_shapes, int row) {
    for (int x = 0; x < 5; x++) {
        // stop searching and return false if we find any clear pixels
        if (stacked_shapes.getPixelValue(x, row) == 0) {
            return false;
        }
    }
    // if we didn't find any clear pixels, this row is contiguous
    return true;
}


/**
 * @brief Given a reference to a MicroBitImage of stacked shapes, remove any
 * contiguous rows of pixels, shifting contents above them down.
 * @returns the number of rows that were eliminated
 */
int eliminate_rows(MicroBitImage& stacked_shapes) {
    // work upwards from the last row
    int row = 4;
    int eliminated = 0;
    do {
        if (is_row_contiquous(stacked_shapes, row)) {
            eliminated++;
            // optimisation: if on the last row, just shift entire image down
            if (row == 4) {
                stacked_shapes.shiftDown(1);
            } else {
                // otherwise, we have to manually drag down all pixels above
                for (int above = row - 1; above >= 0; above--) {
                    for (int x = 0; x < 5; x++) {
                        int value = stacked_shapes.getPixelValue(x, above);
                        stacked_shapes.setPixelValue(x, above + 1, value);
                    }
                }
                // clear the top row, the manual shift down didn't clear it
                for (int x = 0; x < 5; x++) {
                    stacked_shapes.setPixelValue(x, 0, 0);
                }
            }
        } else {
            /*
             * only go up the image when the current row is no longer contiguous
             * --we do this because if we have just eliminated a row, we need
             * to check that same row again because everything will have shifted
             * down by one and the next row (the new row now in this place)
             * might also be full and can be eliminated.
             */
            row--;
        }
    } while (row > 0);
    /*
     * these checks left out the top row, so manually check it now as a special
     * case as the clearing behaviour for it is different
     */
    if (is_row_contiquous(stacked_shapes, 0)) {
        eliminated++;
        for (int x = 0; x < 5; x++) {
            stacked_shapes.setPixelValue(x, 0, 0);
        }
    }
    // return the number of eliminated rows
    return eliminated;
}


void start_new_game() {
    game_over = false; // we're starting a new game
    score = 0;
    // clear the stacked shapes image
    stacked_shapes.clear();
    // // set up event handlers
    ubit.messageBus.listen(
        MICROBIT_ID_BUTTON_A, MICROBIT_BUTTON_EVT_CLICK, rotate_acw
    );
    ubit.messageBus.listen(
        MICROBIT_ID_BUTTON_B, MICROBIT_BUTTON_EVT_CLICK, rotate_cw
    );
    unsigned long stopwatch = ubit.systemTime();
    unsigned long move_speed = 750; // move down one block every 750ms
    // this is the main game loop
    while (not game_over) {
        // generate a new Block
        Block block;
        // starting position
        Point origin(1, -3);
        // this is the down unit-vector
        Vector down(0, 1);
        // if Block can go down
        if (can_shape_move(block, origin, stacked_shapes, down)) {
            /*
             * this is the Block scrolling down loop
             * its exit condition is not checked on every iteration but at a
             * fixed time interval (every time the Block is meant to go down)
             */
            while (true) {
                // allow the player to rotate the Block
                if (rotate_command != 0) {
                    /*
                     * using modulo in these rotation calculations allows
                     * directions to wrap around correctly easily
                     */
                    RotationDirection face;
                    switch (rotate_command) {
                    case -1: // Anticlockwise
                        face = (RotationDirection)(
                            ((unsigned int)block.facing - 1u) % 4u
                        );
                        break;
                    case +1: // Clockwise
                        face = (RotationDirection)(
                            ((unsigned int)block.facing + 1u) % 4u
                        );
                        break;
                    }
                    // check collisions and face this direction if none
                    if (can_shape_rotate(block, origin, stacked_shapes, face)) {
                        block.facing = face;
                    }
                    // reset rotate command to prevent it sticking
                    rotate_command = 0;
                }
                // allow player to shift the Block left or right
                /*
                 * read accelerometer manually for more controlled movement
                 * than what is available from gesture events
                 */
                if (ubit.accelerometer.getX() <= -300) {
                    move_command = -1;
                }
                if (ubit.accelerometer.getX() >= +300) {
                    move_command = +1;
                }
                // take note of any move requests from the player
                if (move_command != 0) {
                    // create unit vector from move command
                    Vector move(move_command, 0);
                    if (can_shape_move(block, origin, stacked_shapes, move)) {
                        origin = origin + move;
                    }
                    // reset move command to prevent it sticking
                    move_command = 0;
                }
                // shift the Block down at a constant rate
                {
                    // check if enough time has elapsed to move the block
                    unsigned long stopwatch_now = ubit.systemTime();
                    if ((stopwatch_now - stopwatch) >= move_speed) {
                        // shift the shape down if it can
                        if (can_shape_move(block, origin, stacked_shapes, down)) {
                            origin = origin + down;
                        } else {
                            // if we can't shift it down, then it's time to stop
                            break;
                        }
                        // update the "stopwatch"
                        stopwatch = stopwatch_now;
                    }
                }
                // draw the scene
                ubit.display.print(stacked_shapes);
                // draw Block with transparency enabled on clear image pixels
                ubit.display.print(block.image(), origin.x, origin.y, 1);
                // delay
                ubit.sleep(10); // 10ms sleep = 100ticks/second
            }
        }
        // if Block Y origin is less than 0 upon not being able to go down
        if (origin.y < 0) {
            // it's Game Over, because the Block could never fully enter screen
            game_over = true;
        } else {
            // the Block has now collided with the stack, copy it to the stack
            stacked_shapes.paste(block.image(), origin.x, origin.y, 1);
            // and remove contiguous rows
            int eliminated_count = eliminate_rows(stacked_shapes);
            // update score when rows are eliminated and blocks stacked
            // scoring: +1 for the block, +5*row count
            score += (1 + eliminated_count * 5);
        }
    }
    // remove event-handlers
    ubit.messageBus.ignore(
        MICROBIT_ID_BUTTON_A, MICROBIT_BUTTON_EVT_CLICK, rotate_acw
    );
    ubit.messageBus.ignore(
        MICROBIT_ID_BUTTON_B, MICROBIT_BUTTON_EVT_CLICK, rotate_cw
    );
    // display score
    ubit.display.scroll("SCORE:");
    ubit.display.scroll(score);
}


int main() {
    // Initialise the micro:bit runtime.
    ubit.init();
    // draw first dot to indicate runtime system is initialised
    ubit.display.image.setPixelValue(0, 2, 255);
    // seed the PRNG with the HWRNG
    ubit.seedRandom();
    // draw second dot to indicate PRNG has been seeded
    ubit.display.image.setPixelValue(2, 2, 255);
    // ensure display is set to black and white mode
    ubit.display.setDisplayMode(DISPLAY_MODE_BLACK_AND_WHITE);
    // draw third and final dot to indicate screen has been set up
    ubit.display.image.setPixelValue(4, 2, 1);
    // introduce the game with a greeting message
    ubit.display.scroll("BLOCKS!");

    // indefinitely start new games
    while (true) {
        start_new_game();
    }

    // TODO: potentially remove this call, if clarified that it is not required.
    // If main exits, there may still be other fibers running or registered event handlers etc.
    // Simply release this fiber, which will mean we enter the scheduler. Worse case, we then
    // sit in the idle task forever, in a power efficient sleep.
    release_fiber();
}
